import json
from pathlib import Path

def report_json(violations: list[dict], ignored_lines: list[dict]=None, ignored_blocks: list[dict]=None, output_path: Path=None):
    """Writes violations to JSON file"""

    # defaults to rolint_results.json in the rolint directory.
    if not output_path:
        output_path = Path("rolint_results.json")
    
    grouped = {}

    # group violations by file and add
    if violations:
        for v in violations:
            file = v["file"]
            grouped.setdefault(file, []).append({
                "line": v["line"],
                "message": f"VIOLATION: {v["message"]}"
            })

    if ignored_lines:
            for flag in ignored_lines:
                file = flag["file"]
                grouped.setdefault(file, []).append({
                    "line": flag["line"],
                    "message": f"FLAG: Ignored or Overridden Line at Line {flag["line"]}"
                })
    
    if ignored_blocks:
            for flag in ignored_blocks:
                file = flag["file"]
                grouped.setdefault(file, []).append({
                    "line": flag["line"],
                    "message": f"FLAG: Ignored or Overridden Code Block at Line {flag["line"]}"
                })
    # Dump output into json
    with open(output_path, "w", encoding="utf-8") as f:
        json.dump(grouped, f, indent=2)




from collections import defaultdict

## Helper file for detecting recursion and mutual recursion in C functions. Recursion is banned through this linter.

def collect_function_definitions(root_node, source_code: bytes) -> dict:
    """
    Walks the AST to collect all function_definition nodes.
    Returns: dict mapping function names to their function_definition node.
    """
    functions = {}

    def walk(node):
        if node.type == "function_definition":
            declarator = node.child_by_field_name("declarator")
            if declarator is not None:
                # Handle nested pointers, declarator chains
                ident_node = declarator
                while ident_node.type != "identifier" and ident_node.child_by_field_name("declarator"):
                    ident_node = ident_node.child_by_field_name("declarator")

                if ident_node.type == "identifier":
                    func_name = source_code[ident_node.start_byte:ident_node.end_byte].decode("utf-8")
                    functions[func_name] = node

        for child in node.children:
            walk(child)

    walk(root_node)
    return functions


def build_call_graph(functions, source_code):
    graph = defaultdict(list)

    for name, body in functions.items():
        if not body:
            continue

        def walk(node):
            if node.type == "call_expression":
                callee_node = node.child_by_field_name("function")
                if callee_node and callee_node.type == "identifier":
                    callee = source_code[callee_node.start_byte:callee_node.end_byte].decode("utf-8")
                    graph[name].append(callee)
            for child in node.children:
                walk(child)

        walk(body)

    return graph

def detect_recursive_functions(call_graph):
    visited = set()
    stack = set()
    recursive = set()

    def dfs(func):
        if func in stack:
            recursive.add(func)
            return
        if func in visited:
            return
        visited.add(func)
        stack.add(func)
        for callee in call_graph.get(func, []):
            dfs(callee)
        stack.remove(func)

    for func in call_graph:
        dfs(func)

    return recursive



## ---------------------------------------- Control Flow Safety Rules ----------------------------------------------------

#Ban unsafe switch statement practices
def check_switch_statement(node, source_code: str) -> list[dict]:
    violations = []

    has_default = False

    for child in node.named_children:
        if child.type == "default_label":
            has_default = True

        def walk_switch_subtree(n):
            nonlocal has_default, violations

            if n.type == "default_label":
                has_default = True

            elif n.type in {"break_statement", "continue_statement"}:
                violations += check_break_continue_in_switch(n, source_code)

            for child in n.children:
                walk_switch_subtree(child)

    walk_switch_subtree(node)

    # Check for fallthrough
    children = [child for child in node.child_by_field_name("body").children if child.type not in {'{', '}'}]
    current_label = None
    current_block = []

    for child in children:
        if child.type in {"case_label", "default_label"}:
            if current_label is not None:
                if not block_has_terminator_or_fallthrough_comment(current_block, source_code):
                    violations.append({
                        "line": current_label.start_point[0] + 1,
                        "message": f"Case '{source_code[current_label.start_byte:current_label.end_byte].decode('utf-8').strip()}' falls through implicitly. Add 'break;', 'return;', or comment like '/* fallthrough */'."
                    })
            current_label = child
            current_block = []
        else:
            current_block.append(child)

    if not has_default:
        violations.append({
            "line": node.start_point[0] + 1,
            "message": "Switch statement missing 'default' case."
        })

    return violations

# Check to make sure we are not using break or continue in standalone switch cases. danger of undefined logic
def check_break_continue_in_switch(node, source_code: str) -> list[dict]:
    violations = []
    current = node.parent

    # Find enclosing control structures
    inside_loop = False
    inside_switch = False

    while current:
        if current.type in {"for_statement", "while_statement", "do_statement"}:
            inside_loop = True
        if current.type == "switch_statement":
            inside_switch = True
            break
        current = current.parent
    if node.type == "break_statement":
        # 'break' is allowed in switch and loop — do not warn
        pass

    return violations


# Helper function to detect fallthrough comment to allow for fallthrough
def block_has_terminator_or_fallthrough_comment(stmts, source_code: str) -> bool:
    for stmt in reversed(stmts):
        text = source_code[stmt.start_byte:stmt.end_byte].decode("utf-8").strip()

        if stmt.type in {"break_statement", "return_statement", "goto_statement", "throw_statement"}:
            return True
        if "fallthrough" in text.lower():
            return True
        if stmt.type != "comment":
            break  # hit a code statement that's not a terminator
    return False


# Ban Recursion (called outside of walk function in main)
def check_recursion(root_node, source_code: bytes) -> list[dict]:
    from rolint.rules.func_analysis_c import (
        collect_function_definitions,
        build_call_graph,
        detect_recursive_functions
    )

    violations = []
    source_str = source_code.decode("utf-8")

    functions = collect_function_definitions(root_node, source_code)
    call_graph = build_call_graph(functions, source_code)
    recursive_funcs = detect_recursive_functions(call_graph)

    for name in recursive_funcs:
        body = functions.get(name)
        if body:
            violations.append({
                "line": body.start_point[0] + 1,
                "message": f"Recursive function '{name}' is banned. Use an iterative alternative."
            })

    return violations



def check_banned_functions(node, source_code: str) -> list[dict]:

    """
    Function checks each node for a function identifier and ensures the function name
    does not match any of the banned functions from the c standard library.
    Returns: list[dict] of violations which include a line, function name, and a message
    """

    banned_functions = {
        "gets", "strcpy", "printf", "sprintf", "vsprintf",
        "strcat", "strncat", "scanf", "sscanf", "fscanf",
        "strtok", "atoi", "atol", "atof", "atoll", "setjmp", "longjmp",
        "malloc", "calloc", "free", "realloc"
    }

    violations = []

    
   
    func_node = node.child_by_field_name("function")

    # If the field is missing, try fallback
    if func_node is None:
        # Try to find the first identifier before the argument list
        for child in node.children:
            if child.type == "identifier":
                func_node = child
                break

    if func_node and func_node.type == "identifier":
        func_name = source_code[func_node.start_byte:func_node.end_byte].decode("utf-8")
        if func_name in banned_functions:
            violations.append({
                "line": node.start_point[0] + 1,
                "function": func_name,
                "message": f"Usage of function '{func_name}' is banned. Please use safer alternative."
            })

    return violations



import re
from typing import Optional, Tuple



PRIMITIVE = {"char","short","int","long","float","double","bool","_Bool"}

def _slice(src: bytes, n) -> str:
    return src[n.start_byte:n.end_byte].decode("utf-8", errors="ignore").strip()

def _unwrap_parens(n):
    while n and n.type == "parenthesized_expression":
        inner = n.child_by_field_name("expression")
        n = inner or (n.children[1] if len(n.children) >= 2 else n)
    return n

def _strip_quals(t: str) -> str:
    return re.sub(r"\b(const|volatile|restrict|signed|register)\b", "", t).strip()

def _normalize_type(t: Optional[str], typedefs: Optional[dict]=None):
    """
    Returns a tuple: (kind, struct_name, ptr_depth)
      kind: one of PRIMITIVE or 'struct' or None
      struct_name: e.g. 'S' if kind=='struct' else None
      ptr_depth: number of '*'
    """
    if not t:
        return None, None, 0
    if typedefs and t in typedefs:
        t = typedefs[t]
    t = _strip_quals(t)
    ptr_depth = t.count("*")
    base = t.replace("*", "").strip()
    if base.startswith("struct "):
        return "struct", base.split(None, 1)[1], ptr_depth
    return base, None, ptr_depth

def _infer_literal_type(node, src: bytes) -> Optional[str]:
    # float literals: contain '.' or exponent (dec/hex); 'f'/'F' suffix ⇒ float else double
    if node.type == "number_literal":
        t = _slice(src, node)
        if any(ch in t for ch in ('.','e','E','p','P')):
            return "float" if re.search(r"[fF]$", t) else "double"
        return "int"
    if node.type == "char_literal":
        return "char"
    return None

def _is_member_access(n) -> bool:
    # TS C uses 'field_expression' for both '.' and '->'. Some grammars also expose 'arrow_expression'.
    return n.type in {"field_expression", "arrow_expression"}

def _field_parts(node):
    # Return (base_expr_node, field_identifier_node)
    base = node.child_by_field_name("argument") or node.child_by_field_name("object")
    if base is None and node.children:
        base = node.children[0]
    field = node.child_by_field_name("field")
    if field is None:
        for c in reversed(node.children):
            if c.type in ("field_identifier","property_identifier","identifier"):
                field = c
                break
    return base, field

def resolve_struct_field_type(expr_node, src: bytes, symtab: dict, struct_table: dict, typedefs: Optional[dict]=None) -> Optional[str]:
    """
    For a.b / a->b return the declared field type string, or None if unknown.
    """
    if not _is_member_access(expr_node):
        return None
    base, field = _field_parts(expr_node)
    if not base or not field:
        return None
    base_t = resolve_expr_type(base, src, symtab, struct_table, typedefs)
    kind, sname, ptr = _normalize_type(base_t, typedefs)
    if kind != "struct" or not sname:
        return None
    # (optional) sanity: '.' expects ptr==0, '->' expects ptr>=1
    field_name = _slice(src, field)
    return struct_table.get(sname, {}).get(field_name)

def resolve_expr_type(node, src: bytes, symtab: dict, struct_table: dict, typedefs: Optional[dict]=None) -> Optional[str]:
    """
    Try to determine the type string of an expression: identifier, member, or literal.
    Returns a textual type (e.g., 'int', 'double', 'struct S', 'struct S *', etc.) or None.
    """
    node = _unwrap_parens(node)
    if node.type == "identifier":
        return symtab.get(_slice(src, node))
    if _is_member_access(node):
        return resolve_struct_field_type(node, src, symtab, struct_table, typedefs)
    lit = _infer_literal_type(node, src)
    if lit:
        return lit
    # (extend later: unary, binary, calls, array subscripts)
    return None

def compose_decl_type(
    base_type_text: str,
    declarator_node,
    source_code: bytes,
) -> Tuple[Optional[str], Optional[str]]:
    """
    Return (var_name, full_type) from a single declarator (handles pointer layers).
    Example outputs: ('a', 'int'), ('p', 'struct S *'), ('pp', 'int **')
    """
    ptr = 0
    d = declarator_node

    # Walk down through pointer/array/function wrappers to reach the identifier.
    while d:
        t = d.type
        if t == "pointer_declarator":
            ptr += 1
            d = d.child_by_field_name("declarator") or (d.children[-1] if d.children else None)
            continue
        if t in {"array_declarator", "function_declarator"}:
            # You can record array/function info later if you want; for now just descend.
            d = d.child_by_field_name("declarator") or (d.children[0] if d.children else None)
            continue
        if t == "parenthesized_declarator":
            d = d.child_by_field_name("declarator") or (d.children[1] if len(d.children) > 1 else None)
            continue
        break

    if not d or d.type != "identifier":
        return None, None

    name = _slice(source_code, d)
    full = (base_type_text + " " + ("*" * ptr)).strip()
    return name, full


## ------------------------------------------- Type Safety Rules ------------------------------------------------

#Check to ensure there are no dangerous type conversions
def check_implicit_conversion_in_declaration(node, source_code: bytes, symbol_table: dict,
                                             struct_table: dict, typedefs: Optional[dict]=None) -> list[dict]:
    violations = []
    if node.type != "declaration":
        return violations

    type_node = node.child_by_field_name("type")
    if not type_node:
        return violations
    base_type = _slice(source_code, type_node)

    # Forward-decl like `struct MSG;` has no init_declarators
    has_declarators = any(ch.type == "init_declarator" for ch in node.named_children)

    for child in node.named_children:
        if child.type != "init_declarator":
            continue

        decl = child.child_by_field_name("declarator")
        value_node = child.child_by_field_name("value")
        if not decl:
            continue

        # derive var name + pointer depth for this declarator
        var_name = None
        ptr = 0
        d = decl
        while d and d.type == "pointer_declarator":
            ptr += 1
            d = d.child_by_field_name("declarator") or (d.children[-1] if d.children else None)
        if d and d.type == "identifier":
            var_name = _slice(source_code, d)
        if not var_name:
            continue

        declared_type_full = (base_type + " " + ("*" * ptr)).strip()
        symbol_table[var_name] = declared_type_full  # per-variable, not just base

        # initializer checks (only for simple numeric literals or resolvable struct fields)
        if value_node:
            rhs_type = resolve_expr_type(value_node, source_code, symbol_table, struct_table, typedefs)
            # only compare when known
            if rhs_type:
                lhs_kind, _, _ = _normalize_type(declared_type_full, typedefs)
                rhs_kind, _, _ = _normalize_type(rhs_type, typedefs)

                if lhs_kind in {"int","short","char"} and rhs_kind in {"float","double"}:
                    violations.append({
                        "line": node.start_point[0] + 1,
                        "message": f"Implicit conversion from {rhs_kind} literal to '{lhs_kind}' in declaration of '{var_name}' may lose precision."
                    })

                if lhs_kind == "char" and value_node.type == "number_literal":
                    txt = _slice(source_code, value_node)
                    if not any(ch in txt.lower() for ch in ('.','e','p')):
                        try:
                            val = int(txt, 0)
                            if val > 127 or val < -128:
                                violations.append({
                                    "line": node.start_point[0] + 1,
                                    "message": f"Value {val} may overflow 'char' in declaration of '{var_name}'."
                                })
                        except ValueError:
                            pass

    return violations


#Check for implicit type conversion when assigning new values to integers
def check_implicit_conversion_in_assignment(node, source_code: bytes, symbol_table: dict,
                                            struct_table: dict, typedefs: Optional[dict]=None) -> list[dict]:
    conversion_table = {
        "double": ["float", "long", "int", "short", "char"],
        "float":  ["int", "short", "char"],
        "long":   ["int", "short", "char"],
        "int":    ["short", "char"],
        "short":  ["char"]
    }
    violations = []
    if node.type != "assignment_expression":
        return violations

    left = _unwrap_parens(node.child_by_field_name("left"))
    right = _unwrap_parens(node.child_by_field_name("right"))
    if not left or not right:
        return violations

    # LHS type: identifier or struct field
    lhs_type = None
    lhs_name = _slice(source_code, left)
    if left.type == "identifier":
        lhs_type = symbol_table.get(lhs_name)
    elif _is_member_access(left):
        lhs_type = resolve_struct_field_type(left, source_code, symbol_table, struct_table, typedefs)
    else:
        return violations

    if not lhs_type:
        return violations

    rhs_type = resolve_expr_type(right, source_code, symbol_table, struct_table, typedefs)

    
    if not rhs_type:
        return violations

    lkind, _, _ = _normalize_type(lhs_type, typedefs)
    rkind, _, _ = _normalize_type(rhs_type, typedefs)

    # float/double → int/short/char
    if lkind in {"int","short","char"} and rkind in {"float","double"}:
        violations.append({
            "line": node.start_point[0] + 1,
            "message": f"Type mismatch or implicit conversion from {rkind} to '{lkind}' in assignment to '{lhs_name}' may lose precision."
        })
    # general narrowing (e.g., long → short)
    elif lkind in conversion_table and rkind in conversion_table[lkind]:
        violations.append({
            "line": node.start_point[0] + 1,
            "message": f"Implicit narrowing conversion from '{rkind}' to '{lkind}' in assignment to '{lhs_name}' may lose precision."
        })

    if lkind == "char" and right.type == "number_literal":
        txt = _slice(source_code, right)
        if not any(ch in txt.lower() for ch in ('.','e','p')):
            try:
                val = int(txt, 0)
                if val > 127 or val < -128:
                    violations.append({
                        "line": node.start_point[0] + 1,
                        "message": f"Value {val} may overflow 'char' in assignment to '{lhs_name}'."
                    })
            except ValueError:
                pass

    return violations

#Casting between pointers and arithmetic types banned.
def check_casting(node, source_code: bytes, symbol_table: dict,
                  struct_table: dict, typedefs: Optional[dict]=None) -> list[dict]:
    violations = []
    if node.type != "cast_expression":
        return violations

    type_node = node.child_by_field_name("type")
    value_node = node.child_by_field_name("value")
    if not type_node or not value_node:
        return violations

    cast_to_text = _slice(source_code, type_node)
    to_kind, to_struct, to_ptr = _normalize_type(cast_to_text, typedefs)


    cast_from_text = resolve_expr_type(value_node, source_code, symbol_table, struct_table, typedefs)
    from_kind, from_struct, from_ptr = _normalize_type(cast_from_text, typedefs)


    if to_kind is None and from_kind is None:
        return violations

    # Pointer <-> arithmetic is banned
    to_is_ptr = to_ptr > 0
    from_is_ptr = from_ptr > 0
    to_is_arith = to_kind in PRIMITIVE
    from_is_arith = from_kind in PRIMITIVE

    # Struct value casts (not pointers) are invalid/banned in practice
    to_is_struct_val = (to_kind == "struct" and to_ptr == 0)
    from_is_struct_val = (from_kind == "struct" and from_ptr == 0)

    if (to_is_ptr and from_is_arith) or (from_is_ptr and to_is_arith):
        violations.append({
            "line": node.start_point[0] + 1,
            "message": f"Cast from '{cast_from_text}' to '{cast_to_text}' between pointer and arithmetic type is banned."
        })
    elif to_is_struct_val or from_is_struct_val:
        violations.append({
            "line": node.start_point[0] + 1,
            "message": f"Cast involving struct value type is banned: '{cast_from_text}' → '{cast_to_text}'."
        })

    return violations

#Ban narrowing casts (float casted to int, long casted to short, etc)
def check_narrowing_casts(node, source_code: bytes, symbol_table: dict,
                          struct_table: dict, typedefs: Optional[dict]=None) -> list[dict]:
    violations = []
    if node.type != "cast_expression":
        return violations

    type_node = node.child_by_field_name("type")
    value_node = node.child_by_field_name("value")
    if not type_node or not value_node:
        return violations

    cast_to_text = _slice(source_code, type_node)
    to_kind, _, to_ptr = _normalize_type(cast_to_text, typedefs)

    cast_from_text = resolve_expr_type(value_node, source_code, symbol_table, struct_table, typedefs)
    from_kind, _, from_ptr = _normalize_type(cast_from_text, typedefs)

    # Only arithmetic (non-pointers) participate in narrowing here
    if to_kind not in PRIMITIVE or from_kind not in PRIMITIVE or to_ptr or from_ptr:
        return violations

    narrowing = {
        "double": ["float", "long", "int", "short", "char"],
        "float":  ["int", "short", "char"],
        "long":   ["int", "short", "char"],
        "int":    ["short", "char"],
        "short":  ["char"]
    }

    if from_kind in narrowing and to_kind in narrowing[from_kind]:
        violations.append({
            "line": node.start_point[0] + 1,
            "message": f"Narrowing cast from '{from_kind}' to '{to_kind}' may lose precision or range."
        })

    return violations


"""
This file is to be used mainly for creating a global struct table between C files. This way, we can track 
structs defined in header files in the subsequent files in which they are used.
"""

def build_struct_table(node, source_code:str) -> list[dict]:

    """
        Walks the AST to build a table of tables essentially.
        {
            "StructName":
                {field1: type},
                {field2: type},
        }
    """
    struct_table = {}

    def walk(n):
        if n.type == "struct_specifier":
            name_node = n.child_by_field_name("name")
            if not name_node:
                return  # Skip unnamed structs

            struct_name = source_code[name_node.start_byte:name_node.end_byte].decode("utf-8")

            field_decls = {}
            for child in n.named_children:
                if child.type == "field_declaration_list":
                    for field in child.named_children:
                        if field.type == "field_declaration":
                            type_node = field.child_by_field_name("type")
                            decl_node = field.child_by_field_name("declarator")
                            if type_node and decl_node:
                                field_type = source_code[type_node.start_byte:type_node.end_byte].decode("utf-8").strip()
                                field_name = source_code[decl_node.start_byte:decl_node.end_byte].decode("utf-8").strip()
                                field_decls[field_name] = field_type

            struct_table[struct_name] = field_decls

        # Recurse on children
        for child in n.children:
            walk(child)

    walk(node)

    return struct_table
    





def walk(node, source_code:str, symbol_table: dict, declared_table: dict, used_table: dict, is_global_var, ignored_lines, ignored_blocks) -> list[dict]:

    violations = []

    if node.start_point[0] in ignored_blocks:
        
        return violations  

    if node.start_point[0] not in ignored_lines:
        # Check for banned functions including new and delete
        if node.type == "call_expression":
            violations += check_banned_funcs(node, source_code)


        # Ban delete and new
        elif node.type == "new_expression":
            violations.append({
                "line": node.start_point[0] + 1,
                "function": "new",
                "message": "Usage of 'new' is banned. Use static or stack allocation instead."
            })


        elif node.type == "delete_expression":
            violations.append({
                "line": node.start_point[0] + 1,
                "function": "delete",
                "message": "Usage of 'delete' is banned. Use RAII or static allocation instead."
            })

        #Check for misuse of switch statements
        elif node.type == "switch_statement":
            violations += check_switch_statement(node, source_code)

        #Ban goto
        elif node.type == "goto":
            violations.append({
                "line": node.start_point[0] + 1,
                "message": "Usage of 'goto' and any uncontrolled jumps are banned."
            })
        
        #Ban function-like macros
        elif node.type == "preproc_function_def":
            violations.append({
                "line": node.start_point[0] + 1,
                "message": "Definition of function-like macros are banned. Consider using 'inline'."
            })
    

        for child in node.children:
            violations += walk(child, source_code, symbol_table, declared_table, used_table, is_global_var,
                               ignored_lines, ignored_blocks)
            

    return violations


# CHECKS


def check_banned_funcs(node, source_code: str) -> list[dict]:
    """
    Ensures function call is not banned. Unsafe functions are defined as those who either dynamically allocate memory or
    have the potential to cause overflow. 
    """

    banned_functions = {
        "malloc", "calloc", "realloc", "free",
        "printf", "sprintf", "scanf", "gets", "fgets",
        "rand", "srand", "time", "clock", "gettimeofday",
        "system", "fork", "exec", "exit",
        "va_start", "va_arg", "va_end",
        "cin", "cout", "cerr"
    }

    violations = []

    function_node = node.child_by_field_name('function')
    if function_node is not None:
        name = source_code[function_node.start_byte:function_node.end_byte].decode("utf-8")

        if '::' in name:
            name = name.split('::')[-1]
        
        
        if name in banned_functions:
            violations.append({
                "line": node.start_point[0] + 1,
                "function": name,
                "message": f"Usage of function '{name}' is banned. Please use safer alternative."
            })
    return violations

def check_switch_statement(node, source_code: str) -> list[dict]:
    violations = []
    has_default = False

    body = node.child_by_field_name("body")
    if body is None:
        return violations

    # Flatten to look at named children (statements in body)
    children = body.named_children
    last_case = None
    case_block_has_exit = False

    for i, child in enumerate(children):
        if child.type == "default_label":
            has_default = True

        elif child.type == "case_label":
            if last_case is not None and not case_block_has_exit:
                # If previous case block ended with no break/return
                violations.append({
                    "line": child.start_point[0] + 1,
                    "message": "Switch case statement has implicit fallthrough. Add 'break;', 'return;', or '[[fallthrough]]'"
                })
            last_case = child
            case_block_has_exit = False

        elif child.type in {"break_statement", "return_statement", "throw_statement"}:
            case_block_has_exit = True

        elif child.type == "continue_statement":
            violations.append({
                "line": child.start_point[0] + 1,
                "message": "Use of 'continue' is banned."
            })

    if not has_default:
        violations.append({
            "line": node.start_point[0] + 1,
            "message": "Switch statement missing 'default' case."
        })

    return violations




import ast
import subprocess
from pathlib import Path

# Got to make this part of the linter a lot cleaner due to the AST support in python

## Python rules implemented:
#      - Static type checking (requiring type hints on variables)
#      - Ban unsafe python functions
#      - Enforce safe threading practices (i.e Graceful Termination)
#      - Enforce PEP8 standards
#      - Enforce runtime assertions (with prebuilt template function rather than assert)




class PyRules(ast.NodeVisitor):
    def __init__(self, source: str, path: Path):
        self.src = source
        self.path = path
        self.violations = []
        self.current_function = None # Context for current function
        self.banned_functions = {
            "eval", "exec", "compile"
        }
        self.banned_full = {
            "pickles"
        }
        self.thread_objs = {} # Map for threads {'started': bool, 'joined': bool}
        self.subprocesses = {} # Map for subprocesses

    def add(self, node, msg):
        """
        Helper function to add violations to the list
        """
        self.violations.append({
            "line": node.lineno,
            "message": msg
        })

    def finalize(self):
        # Check to ensure all threading.Threads have been joined
        for tname, data in self.thread_objs.items():
            if data["started"] and not data["joined"]:
                self.add(self.path, f"Thead '{tname}' started but never joined.")
        #subprocess graceful termination check
        for pname, data in self.subprocesses.items():
            if not (data["waited"] or data["terminated"] or data["communicated"]):
                self.add(self.path, f"Subprocess '{pname}' started with Popen but not safely terminated or awaited.")

    # Rule for static type checking
    def visit_FunctionDef(self, node):
        self.current_function = node

        args = node.args.args
        if args and args[0].arg == "self":
            args = args[1:]

        for arg in args + node.args.kwonlyargs:
            #Searching arguments for type annotations
            if arg.annotation is None:
                self.add(arg, f"Paramater '{arg.arg}' missing type annotation. Please specify type.")
        if node.returns is None:
            # Check to make sure return type is declared
            self.add(node, "Function is missing return type annotation.")
        self.generic_visit(node)
        self.current_function = None
    
    # Chillen here because AnnAssign means there's a type hint
    def visit_AnnAssign(self, node: ast.AnnAssign):
        self.generic_visit(node)
    
    #Danger here because a regular assign means no type hint
    def visit_Assign(self, node: ast.Assign):
        if isinstance(node.targets[0], ast.Name) and self.current_function is None:
            self.add(node, f"Variable '{node.targets[0].id}' needs a static type hint.")
        self.generic_visit(node)

    def visit_Call(self, node: ast.Call):
        if isinstance(node.func, ast.Name):
            if node.func.id in self.banned_functions:
                self.add(node, f"Usage of banned function '{node.func.id}'.")
        if isinstance(node.func, ast.Attribute):
            value = node.func.value
            if isinstance(value, ast.Name):
                tup = (value.id, node.func.attr)
                if tup in self.banned_full:
                    self.add(node, f"Banned call '{value.id}.{node.func.attr}()'.")

        if (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.attr == "Thread"
            and node.func.value.id == "threading"
        ):
            # get variable it's assigned to
            parent = getattr(node, 'parent', None)
            if isinstance(parent, ast.Assign) and isinstance(parent.targets[0], ast.Name):
                tname = parent.targets[0].id
                self.thread_objs[tname] = {"started": False, "joined": False}

        #Subprocess tracker
        if (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == "subprocess"
            and node.func.attr == "Popen"
        ):
            parent = getattr(node, 'parent', None)
            if isinstance(parent, ast.Assign) and isinstance(parent.targets[0], ast.Name):
                pname = parent.targets[0].id
                self.subprocesses[pname] = {
                    "waited": False, "terminated": False, "communicated": False
                }

        # start()/join()/terminate()/wait()/communicate() tracking
        if isinstance(node.func, ast.Attribute) and isinstance(node.func.value, ast.Name):

            tname = node.func.value.id

            if tname in self.thread_objs:
                if node.func.attr == "start":
                    self.thread_objs[tname]["started"] = True
                elif node.func.attr == "join":
                    self.thread_objs[tname]["joined"] = True

        if isinstance(node.func, ast.Attribute) and isinstance(node.func.value, ast.Name):
            pname = node.func.value.id
            if pname in self.subprocesses:
                if node.func.attr == "wait":
                    self.subprocesses[pname]["waited"] = True
                elif node.func.attr == "terminate":
                    self.subprocesses[pname]["terminated"] = True
                elif node.func.attr == "communicate":
                    self.subprocesses[pname]["communicated"] = True
    
        
    



def run_python_linter(path: Path, ignored_lines, ignored_blocks) -> list[dict]:
    text = path.read_text(encoding="utf-8")
    try:
        tree = ast.parse(text, filename=str(path))
    except SyntaxError as e:
        return [{
            "line": e.lineno,
            "message": f"Syntax error: {e.msg} at line {e.lineno}. Please resolve before continuing."
        }]

    linter = PyRules(text, path)
    linter.visit(tree) # Visit all nodes (visit function defined in ast.NodeVisitor class)
    linter.finalize() # Finalize the search by ensuring all threads and subprocesses created were properly terminated

    #Run flake8 for PEP8 standards
    result = subprocess.run(
        ["flake8", str(path), "--ignore=E501,E302,E231,E262,E261,W291,W292"], 
        capture_output=True, text=True
    )
    for line in result.stdout.splitlines():
        parts = line.split(":")
        if len(parts) >= 4:
            lnum = int(parts[1])
            msg = ":".join(parts[3:]).strip()
            linter.violations.append({"line": lnum, 
                                      "message": f"PEP8 Violation: {msg}"})

    ignored_line_ranges = get_block_ranges(tree, ignored_blocks)

    def is_ignored(line):
        # Helper function to determine if a line is ignored or not
        if line in ignored_lines:
            return True
        for start, end in ignored_line_ranges:
            if start <= line <= end:
                return True
        return False

    # Filter out ignored violations
    linter.violations = [
        v for v in linter.violations if not is_ignored(v["line"])
    ]

    return linter.violations



def get_block_ranges(tree, ignored_lines):
    """
    Helper to determine a coding block's size to ignore (for rolint: ignore-block command)
    """

    ignored_ranges = []

    for node in ast.walk(tree):
        # Skip if the node doesn't have lineno (like `Load`, `Store`, etc.)
        if hasattr(node, "lineno") and node.lineno in ignored_lines:
            # Estimate block range using lineno and end_lineno (Python 3.8+)
            start = node.lineno
            end = getattr(node, "end_lineno", None)

            if end is None:
                # fallback for Python <3.8 or missing info
                # this is very rough, assumes block is ~1 line
                end = node.lineno + 1

            ignored_ranges.append((start, end))
    
    return ignored_ranges


from .c_rule_functions.check_banned_funcs import check_banned_functions
from .c_rule_functions import function_and_vars_use
from .c_rule_functions import control_flow_rules
from .c_rule_functions import type_safety

"""
    All C Rules will be encoded in this script here. This will consist of a set of banned functions,
    unsafe practices, etc.
"""

def walk(node, source_code:str, symbol_table: dict, declared_table: dict, used_table: dict, struct_table, is_global_var, ignored_lines, ignored_blocks) -> list[dict]:

    
    violations = []

    if (node.start_point[0] in ignored_blocks):
        
        return violations  

    if node.start_point[0] not in ignored_lines:

        if node.type == "call_expression":
            #Check expression call to see if function is banned
            violations += check_banned_functions(node, source_code)
            
            #Check for side effects in function call
            violations += function_and_vars_use.check_side_effects_in_func_call(node, source_code)

        elif node.type == "declaration":  

            if is_global_var:
                violations += function_and_vars_use.check_global(node, source_code)

            #Check declarations rules (multiple conversions, no initialization)  
            violations += function_and_vars_use.check_declaration(node, source_code)


            #Check type conversions
            violations += type_safety.check_implicit_conversion_in_declaration(node, source_code, symbol_table, struct_table)

            # Track declared vars
            for child in node.named_children:
                if child.type == "init_declarator":
                    ident = child.child_by_field_name("declarator")
                    if ident and ident.type == "identifier":
                        name = source_code[ident.start_byte:ident.end_byte].decode("utf-8").strip()
                        declared_table["variables"][name] = node.start_point[0] + 1

        elif node.type == "assignment_expression":
            violations += type_safety.check_implicit_conversion_in_assignment(node, source_code, symbol_table, struct_table)

        elif node.type == "cast_expression":
            violations += type_safety.check_casting(node, source_code, symbol_table, struct_table)
            violations += type_safety.check_narrowing_casts(node, source_code, symbol_table, struct_table)
        
        elif node.type == "continue_statement":
            violations.append({
                "line": node.start_point[0] + 1,
                "message": "Use of 'continue' is banned."
            })
        elif node.type == "goto_statement":
            #Specifically banning goto statements
            violations.append({
                "line": node.start_point[0] + 1,
                "message": "Usage of 'goto' is banned. Please use structured control flow logic."
            })
        
        elif node.type == "switch_statement":
            violations += control_flow_rules.check_switch_statement(node, source_code)
        elif node.type == "preproc_function_def":
            violations += check_function_like_macros(node, source_code)

        ##Checks for unused funcs or vars
        elif node.type == "function_definition":
            is_global_var = False
            func_node = node.child_by_field_name("declarator")
            if func_node:
                ident_node = func_node.child_by_field_name("declarator")
                if ident_node and ident_node.type == "identifier":
                    name = source_code[ident_node.start_byte:ident_node.end_byte].decode("utf-8").strip()
                    declared_table["functions"][name] = node.start_point[0] + 1

        elif node.type == "identifier":

            name = source_code[node.start_byte:node.end_byte].decode("utf-8").strip()

            # Only mark as used if not part of a declaration
            parent = node.parent
            if parent and parent.type != "init_declarator" and parent.type != "declaration":
                if name in declared_table["variables"]:
                    used_table["variables"].add(name)
                if name in declared_table["functions"]:
                    used_table["functions"].add(name) 

    if node.start_point[0] not in ignored_blocks:
        for child in node.children:
            violations += walk(child, source_code, symbol_table, declared_table, used_table, struct_table, is_global_var, ignored_lines, ignored_blocks)
        

    return violations



## ------------------------------------- Function / Variable Use Rules -----------------------------------------



#Check for function like macros
def check_function_like_macros(node, source_code: str) -> list[dict]:
    violations = []

    name_node = node.child_by_field_name("name")
    if name_node:
        name = source_code[name_node.start_byte:name_node.end_byte].decode("utf-8")
        violations.append({
            "line": node.start_point[0] + 1,
            "message": f"Function-like macro '{name}' detected. Usage of function-like macros is banned. Use inline functions instead."
        })

    return violations

 
## ----------------------------------------------- HEADER RULES -----------------------------------------------------------

# Ensure header guards
def check_header_guard(source_code: bytes, file_path: str) -> list[dict]:
    lines = source_code.splitlines()
    violations = []

    guard_macro = None

    # Look for #ifndef and matching #define in the first 10 lines
    for i, line in enumerate(lines):
        stripped = line.strip().decode("utf-8")  #  decode from bytes to str
        if stripped.startswith("#ifndef"):
            parts = stripped.split()
            if len(parts) == 2:
                guard_macro = parts[1]

        elif stripped.startswith("#define") and guard_macro:
            if guard_macro not in stripped:
                guard_macro = None  # Reset if mismatch
            else:
                break

    # Look for #endif near end
    has_endif = any(
        line.strip().decode("utf-8").startswith("#endif") for line in lines[-10:]
    )

    if not (guard_macro and has_endif):
        violations.append({
            "line": 1,
            "message": f"Header file '{file_path}' is missing proper include guards (#ifndef / #define / #endif)."
        })

    return violations


def check_object_definitions_in_header(tree, source_code: str) -> list[dict]:
    violations = []

    def is_function_declaration(node):
        return node.type == "function_declaration"

    def walk(n):
        nonlocal violations
        if n.type == "declaration":
            # Check if this is a definition (i.e., has init_declarator with value)
            for child in n.named_children:
                if child.type == "init_declarator":
                    value = child.child_by_field_name("value")
                    if value is not None:
                        ident = child.child_by_field_name("declarator")
                        if ident:
                            name = source_code[ident.start_byte:ident.end_byte].decode("utf-8")
                            violations.append({
                                "line": n.start_point[0] + 1,
                                "message": f"Object '{name}' defined in header file. Object definitions are banned in headers."
                            })
        for child in n.children:
            walk(child)

    walk(tree.root_node)
    return violations


# Check to make sure everything that is declared is used
def check_unused(declared_symbols, used_symbols):
    violations = []

    unused_vars = set(declared_symbols["variables"].keys()) - used_symbols["variables"]
    for name in unused_vars:
        line = declared_symbols["variables"][name]
        violations.append({
            "line": line,
            "message": f"Variable '{name}' declared but never used."
        })

    unused_funcs = set(declared_symbols["functions"].keys()) - used_symbols["functions"]
    for name in unused_funcs:
        line = declared_symbols["functions"][name]
        violations.append({
            "line": line,
            "message": f"Function '{name}' defined but never called."
        })

    return violations

# Ban Recursion (called outside of walk function in main)
def check_recursion(root_node, source_code: bytes) -> list[dict]:
    from rolint.rules.func_analysis_c import (
        collect_function_definitions,
        build_call_graph,
        detect_recursive_functions
    )

    violations = []
    source_str = source_code.decode("utf-8")

    functions = collect_function_definitions(root_node, source_code)
    call_graph = build_call_graph(functions, source_code)
    recursive_funcs = detect_recursive_functions(call_graph)

    for name in recursive_funcs:
        body = functions.get(name)
        if body:
            violations.append({
                "line": body.start_point[0] + 1,
                "message": f"Recursive function '{name}' is banned. Use an iterative alternative."
            })

    return violations




## Control Flow Safety Rules
# - No Goto <-- DONE
# - no break; continue inside switch statements <-- DONE
# - all switch statements must have a default <-- DONE
# - No recursion <-- DONE

## Memory Safety
# - No malloc, calloc, or free statements (dynamic memory allocation not allowed) <-- DONE 
# - No use of NULL without type context 
# - No object definitions in header files <-- DONE

## Function and Variable Use 
# - No unused variables / functions <-- DONE
# - No global variables unless const <-- DONE
# - No side effects in function arguments <-- DONE  




import subprocess
import threading
import pickle

# Missing type hint for variable
my_data = "hello"

# Unsafe function
eval("print('hi')")

# Unsafe module call
pickle.loads(b"bad_data")

# Thread created but never joined
def thread_example():
    t = threading.Thread(target=lambda: print("Running"))
    t.start()

# Function missing annotations
# rolint: ignore
def do_work(x, y):
    return x + y

# Bare assert
assert 2 + 2 == 4

# Another bare assert as a call
def check():
    assert 1 == 1

# Subprocess without termination
def run_proc():
    proc = subprocess.Popen(["ls", "-l"])
    print("Started process")



from pathlib import Path
from collections import defaultdict
from rolint.parser import parser as parser_module
from rolint.rules import c_rules
from rolint.rules.python_rules import run_python_linter
from rolint.rules import cpp_rules
import sys
from rolint.rules import override
from rolint.reporter.json import report_json
from rolint.rules.struct_table_builder import build_struct_table


EXTENSION_MAP = {
    ".c": "c",
    ".h": "c",
    ".cpp": "cpp",
    ".cc": "cpp",
    ".cxx": "cpp",
    ".hpp": "cpp",
    ".hxx": "cpp",
    ".py": "python"
}

def detect_language(path: Path) -> str:
    return EXTENSION_MAP.get(path.suffix.lower(), None)

def collect_files(base_path: Path) -> dict[str, list[Path]]:
    lang_files = defaultdict(list)
    for file_path in base_path.rglob("*"):
        if file_path.is_file():
            lang = detect_language(file_path)
            if lang:
                lang_files[lang].append(file_path)
    return lang_files

def run_linter(path: Path, lang: str = None, output_format: str = "text", output_path: Path = None):

    violations = []
    ignored_lines = []
    ignored_blocks = []
    if path.is_dir():
        lang_to_files = collect_files(path)
        if not lang_to_files:
            print("ERROR: No source files found.")
            return
        for lang, files in lang_to_files.items():
            print(f"\n🔧 Linting {len(files)} {lang.upper()} file(s):")
            for f in files:
                print(f"  - {f}")
                
                v, il, ib = run_file_lint(f, lang)
                violations.extend(v)
                ignored_lines.extend(il or [])
                ignored_blocks.extend(ib or [])
        
        #Exit with status 1 code if there are violations to prevent commit
        if violations:
            if output_format == "json":
                report_json(violations, ignored_lines, ignored_blocks, output_path=output_path)
                print(f"📄 Output at {output_path}")
            print("Blocking Commit.")
            print(" - John 8:11")
            sys.exit(1)
        else:
            sys.exit(0)

    elif path.is_file():
        inferred_lang = lang or detect_language(path)
        if not inferred_lang:
            print(f"⚠️ Could not detect language for {path}")
            return
        print(f"🔍 Linting: {path}")
        print(f"🌐 Language: {inferred_lang}")
        print(f"📤 Output format: {output_format}")
        run_file_lint(path, inferred_lang)

        if violations:
            if output_format == "json":   
                report_json(violations, ignored_lines, ignored_blocks, output_path)
                print(f"📄 Output at {output_path}")
            print("Blocking commit.")
            print(" - John 8:11")
            sys.exit(1)
        else:
            sys.exit(0)
    else:
        print(f"❌ Path does not exist: {path}")


def run_file_lint(file_path: Path, lang: str):
    violations = []
    ignored_lines = []
    ignored_blocks = []


    if lang == "c":
        tree, source = parser_module.parse_file(file_path, lang)
        ## Tables for tracking variable contexts.
        symbol_table = {}
        declared_table = {
            "variables": {}, 
            "functions": {}
        }
        used_table = {
            "variables": set(),
            "functions": set()
        }
        global_struct_table = build_struct_table(tree.root_node, source)

        ignored_lines, ignored_blocks = override.detect_override_lines(source)

        if file_path.suffix in {".h"}:
            violations += c_rules.check_header_guard(source, str(file_path))
            violations += c_rules.check_object_definitions_in_header(tree, source)
        else:
            violations += c_rules.walk(tree.root_node, source, symbol_table, declared_table, used_table, global_struct_table,
                                        is_global_var=True, ignored_lines=ignored_lines, ignored_blocks=ignored_blocks)
            violations += c_rules.check_recursion(tree.root_node, source)
            violations += c_rules.check_unused(declared_table, used_table)
        
        if violations:
            for v in violations:
                v["file"] = str(file_path)
                print(f"🚫 VIOLATION: {file_path}:{v['line']} - {v['message']}")

        if ignored_blocks or ignored_lines:
            for il in ignored_lines:
                il["file"] = str(file_path)
                print(f"⚠️  FLAG: {file_path} - Line ignored or overridden at line {il['line']}")
            for il in ignored_blocks:
                il["file"] = str(file_path)
                print(f"⚠️  FLAG: {file_path} - Line ignored or overridden at block {il['line']}")

        
        

    elif lang in {"cpp"}:
        tree, source = parser_module.parse_file(file_path, lang)
        ## Tables for tracking variable contexts.
        symbol_table = {}
        declared_table = {
            "variables": {}, 
            "functions": {}
        }
        used_table = {
            "variables": set(),
            "functions": set()
        }

        ignored_lines, ignored_blocks = override.detect_override_lines(source)

        violations += cpp_rules.walk(tree.root_node, source, symbol_table, declared_table, used_table, is_global_var=True,
                                     ignored_lines=ignored_lines, ignored_blocks=ignored_blocks)

        if violations:
            for v in violations:
                v["file"] = str(file_path)
                print(f"🚫 VIOLATION: {file_path}:{v['line']} - {v['message']}")

        if ignored_blocks or ignored_lines:
            for il in ignored_lines:
                il["file"] = str(file_path)
                print(f"⚠️  FLAG: {file_path} - Line ignored or overridden at line {il['line']}")
            for il in ignored_blocks:
                il["file"] = str(file_path)
                print(f"⚠️  FLAG: {file_path} - Line ignored or overridden at block {il['line']}")


    elif lang in {"python"}:
        source = file_path.read_text(encoding="utf-8")
        ignored_lines, ignored_blocks = override.detect_py_overrides(source)
        

        violations += run_python_linter(file_path, ignored_lines, ignored_blocks)
        
        if violations:
            for v in violations:
                v["file"] = str(file_path)
                print(f"🚫 VIOLATION: {file_path}:{v['line']}: {v['message']}")
        
        if ignored_blocks or ignored_lines:
            for il in ignored_lines:
                il["file"] = str(file_path)
                print(f"⚠️  FLAG: {file_path} - Line ignored or overridden at line {il['line']}")
            for il in ignored_blocks:
                il["file"] = str(file_path)
                print(f"⚠️  FLAG: {file_path} - Line ignored or overridden at block {il['line']}")
    else:
        print(f"⚠️ Unknown language: {lang}")
    

    return violations, ignored_lines, ignored_blocks



import typer
from pathlib import Path
from rolint.config import load_config, save_config
from rolint.main import run_linter

##Main entry point for command line with rolint
app = typer.Typer(help="Rolint - Linter for robotics code (C, C++, Python)")

@app.command()
def check(
    path: Path = typer.Argument(..., help="Path to folder or file to lint"),
    lang: str = typer.Option(None, "--lang", "-l", help="Optional: Force language (c | cpp | python)"),
    output: str = typer.Option("text", "--output", "-o", help="Output format: text | json"),
    output_path: Path = typer.Option(None, "--output-path", "-p", help="Optional: Path to write JSON output"),
):
    """
    Run safety checks on code.

    Examples: \n
      rolint check [OPTIONS][PATH] checks file[s] at specified path \n
    """

    config = load_config()
    output_path_str = output_path or config.get("output_path") or "rolint_results.json"
    true_output_path = Path(output_path_str)
    run_linter(path, lang=lang, output_format=output, output_path=true_output_path)


@app.command()
def set_config(
    output_path: str = typer.Option(..., "--output-path", "-p", help="Set default output path for JSON reports")
):
    """
    Set persistent configuration options for rolint. \n

    """
    config = load_config()
    config["output_path"] = output_path
    save_config(config)
    typer.echo(f"✅ Output path set to: {output_path}")

@app.command()
def show_config():
    """
    Show current configuartion settings.
    """
    from rolint.config import load_config
    config = load_config()
    typer.echo("Current configuration:")
    for key, value in config.items():
        typer.echo(f"  {key}: {value}")


import json
from pathlib import Path

CONFIG_FILE = Path.home() / ".rolint_config.json"

# Set up default configuration for users
DEFAULT_CONFIG = {
    "output_path": "rolint_results.json"
}

# load config file
def load_config() -> dict:
    if CONFIG_FILE.exists():
        try:
            with open(CONFIG_FILE, "r", encoding="utf-8") as f:
                return json.load(f)
        except json.JSONDecodeError:
            print("Config file may be corrupted. Loading default configuration settings.")
            # If config file is corrupted, return default
            return DEFAULT_CONFIG.copy()
    return DEFAULT_CONFIG.copy() 


# save new configuration to config file
def save_config(config: dict) -> dict:
    with open(CONFIG_FILE, "w", encoding="utf-8") as f:
        json.dump(config, f, indent=2)


def detect_override_lines(source_code: str) -> set[int]:
    """
    Parses the source and returns a set of line numbers that should be ignored.
    """

    source_code = source_code.decode('utf-8', errors='ignore')


    ignored_lines = []
    ignored_blocks = []
    lines = source_code.splitlines()
    for i, line in enumerate(lines):
        if "rolint: ignore" in line and "rolint: ignore-block" not in line:
            ignored_lines.append({
                        "line": i + 1
                    })  
        if "rolint: ignore-block" in line:
            ignored_blocks.append({
                        "line": i + 1
                    })

    return ignored_lines, ignored_blocks

def detect_py_overrides(source_code: str) -> tuple[list[int], list[int]]:
    ignored_lines = []
    ignored_blocks = []


    for lineno, line in enumerate(source_code.splitlines(), start=1):
        stripped = line.strip()

        if "rolint: ignore-block" in stripped:
           ignored_blocks.append({
                        "line": lineno + 1
                    })
        elif "rolint: ignore" in stripped and "rolint: ignore-block" not in stripped:
            ignored_lines.append({
                        "line": lineno + 1
                    })

    return ignored_lines, ignored_blocks



# Check if a global var is marked as a constant or volatile
def check_global(node, source_code: str) -> list[dict]:
    violations = []


    # Gather all modifier keywords to check for const or volatile
    modifiers = set()
    for child in node.children:
        if child.type == "type_qualifier" or child.type == "storage_class_specifier":
            qualifier = source_code[child.start_byte:child.end_byte].decode("utf-8")
            modifiers.add(qualifier)

    is_cleared = "const" in modifiers or "volatile" in modifiers or "static" in modifiers or "extern" in modifiers

    # If it doesn't have const or volatile, extract the node info and add to violations
    if not is_cleared:
        for child in node.named_children:
            ident = None
            if child.type == "init_declarator":
                ident = child.child_by_field_name("declarator")
            elif child.type == "array_declarator":
                ident = child.child_by_field_name("declarator")

            while ident and ident.type in {"pointer_declarator", "array_declarator"}:
                ident = ident.child_by_field_name("declarator")

            if ident and ident.type == "identifier":
                var_name = source_code[ident.start_byte:ident.end_byte].decode("utf-8")
                violations.append({
                    "line": node.start_point[0] + 1,
                    "message": f"Global variable '{var_name}' must be marked 'const', 'extern', 'static', or 'volatile'."
                })

    return violations

#Check declaration function (Functions and Variable use)
def check_declaration(node, source_code: str) -> list[dict]:
    """
    Function to check variable declarations. Enforces two rules: multiple declarations on one line and unitialized variables.
    Returns a list of violations which includes a line, variable name, and a message.
    """


    violations = []

    if node.type != "declaration":
        return violations

    # Extract all relevant declarators
    declarators = [
        child for child in node.named_children
        if child.type in {"init_declarator", "identifier"}
    ]

    # Check for multiple variables declared
    if len(declarators) > 1:
        vars_declared = []
        for declared in declarators:
            if declared.type == "identifier":
                var_name = source_code[declared.start_byte:declared.end_byte].decode("utf-8").strip()
                vars_declared.append(var_name)
            elif declared.type == "init_declarator":
                ident = declared.child_by_field_name("declarator")
                if ident and ident.type == "pointer_declarator":
                    ident = ident.child_by_field_name("declarator")
                if ident and ident.type == "identifier":
                    var_name = source_code[ident.start_byte:ident.end_byte].decode("utf-8").strip()
                    vars_declared.append(var_name)
        violations.append({
            "line": node.start_point[0] + 1,
            "message": f"Multiple variables declared in one statement: {', '.join(vars_declared)}. Please separate onto separate lines."
        })
        return violations  # Stop here if multiple declared

    # Check for initialization
    if len(declarators) == 1:
        declared = declarators[0]

        # raw identifier like `int a;` for uninitialized variables
        if declared.type == "identifier":
            var_name = source_code[declared.start_byte:declared.end_byte].decode("utf-8").strip()
            violations.append({
                "line": node.start_point[0] + 1,
                "message": f"Variable '{var_name}' declared without initialization."
            })
            return violations

        # Check to make sure the variable is initialized even if there is an init_declarator
        has_initializer = len(declared.named_children) > 1
        if not has_initializer:
            ident = declared.child_by_field_name("declarator")
            if ident and ident.type == "pointer_declarator":
                ident = ident.child_by_field_name("declarator")
            if ident and ident.type == "identifier":
                var_name = source_code[ident.start_byte:ident.end_byte].decode("utf-8").strip()
            else:
                var_name = "unknown"

            violations.append({
                "line": node.start_point[0] + 1,
                "message": f"Variable '{var_name}' declared without initialization."
            })

    return violations

# Ban side effects in function calls (ex. printf(x++) or printf(getchar()) is unallowed)
def check_side_effects_in_func_call(node, source_code:str) -> list[dict]:
    violations = []

    #Functions that I will allow as I know they don't have side effects to data
    known_pure_functions = {"abs", "sqrt", "strlen", "toupper", "tolower"}

    args_node = node.child_by_field_name("arguments")
    if not args_node:
        return violations
    
    # Recursive walk through all node's children to ensure no side effects
    def contains_side_effects(n):
        if n.type in {"assignment_expression", "update_expression"}:
            return True
        elif n.type == "call_expression":
            func_name_node = n.child_by_field_name("function")
            if func_name_node:
                func_name = source_code[func_name_node.start_byte:func_name_node.end_byte].decode("utf-8")
                if func_name in known_pure_functions:
                    return False 
            return True
        for child in n.children:
            if contains_side_effects(child):
                return True
        return False
    
    for arg in args_node.named_children:
        if contains_side_effects(arg):
            func_node = node.child_by_field_name("function")
            func_name = source_code[func_node.start_byte:func_node.end_byte].decode("utf-8") if func_node else "unknown"
        
            violations.append({
                "line": func_node.start_point[0] + 1,
                "message": f"Side effect or function call in arguments for function call '{func_name}'."
            })


    return violations




##For parsing C and C++ files

from pathlib import Path
import tree_sitter_c as tsc
import tree_sitter_cpp as tscpp
from tree_sitter import Language, Parser

# Supported languages
SUPPORTED_LANGUAGES = {"c", "cpp"}

# Language mappings
LANGUAGE_MAP = {
    "c": Language(tsc.language()),
    "cpp": Language(tscpp.language())
}

# Cache for performance
parsers = {}
for lang in SUPPORTED_LANGUAGES:
    parser = Parser(LANGUAGE_MAP[lang])
    parsers[lang] = parser

def parse_file(file_path: Path, lang: str):
    """
    Parse a file and return (tree, source_code)
    """
    if lang not in SUPPORTED_LANGUAGES:
        raise ValueError(f"Unsupported language: {lang}")

    source_code = file_path.read_bytes() 
    tree = parsers[lang].parse(source_code)
    return tree, source_code
